SystemOrganization addCategory: #MetaInterpreter!

TestCase subclass: #TestInterpreter
	instanceVariableNames: 'anIv'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:50'!
estOnInteractionWithFromDebuggingSupport

	|object collection|

	object := Object new.
	collection := OrderedCollection new.
	
	RunTimeDebuggingSupport onInteractionFor: object with: collection do:  '[:r1 :r2 | 
		r1 trigger halt.
	]'.

	RunTimeDebuggingSupport enableMetaBehaviorForRegistered.
	
	collection add: object.
	
	RunTimeDebuggingSupport removeMetaBehaviorForRegistered.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:13'!
passAsArg: t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:39'!
store: t

	anIv := t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 17:31'!
testCompositeHookOnNew

Interpreter on: MessageReceived for: Point do: (Compiler evaluate: [:reification :reflectogram | 
	reification message selector = #new ifTrue: [
		reification trigger halt.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).	
	]	
] asString).

RunTimeDebuggingSupport evaluate: 'TestClass run'.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:46'!
testDerivativeHooksWithHalt

	"press proceed for the test to finish"

	|t1 t2|
	
	"register halt on creation"
	Interpreter on: MessageReceived for: TestObject do: [:reifications :reflectogram | 
		reifications message selector = #new ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: TestObject.
	
	t1 := TestObject new. "should halt"
	t2 := TestObject new. "should halt."
	
	Interpreter removeMetaBehaviorOf: TestObject.
	
	"register halt on object in invoke -- this can be done either by registering the argument or the receiver"
	Interpreter on: AsArgument for: t2 do: [:reifications :reflectogram |
			reifications receiver = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t2.
	
	t1 testPassingArgument: t2. "should halt"

	Interpreter removeMetaBehaviorOf: t2.
	
	"register halt on interaction"	
	Interpreter on: MessageReceived for: t1 do: [:reifications :reflectogram | 
		(reifications message arguments includes: t2) ifTrue: [self halt].
	].

	Interpreter on: AsArgument for: t1 do: [:reifications :reflectogram | 
		reifications receiver = t2 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t1.


	t1 testInteraction: t2. "should halt because t1 invokes a method of t2"
	t2 testInteraction: t1. "should halt because t2 invokes a method of t1"
	
	Interpreter disableMetaBehaviorFor: t1.
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/5/2013 12:04'!
testHaltOnlyInterpeter

	|t |
	
	HaltInterpreter initialize.
			
	t := TestObject new.
	
	HaltInterpreter on: MessageReceived for: t do: [].
	

	HaltInterpreter enableMetaBehaviorFor: t.
	
	t class.
			
		"Interpreter disableMetaBehaviorFor: t."
	HaltInterpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	Interpreter  initialize.
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:45'!
testHookArg

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:45'!
testHookRead

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivReadCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookReceive

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookSend

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: sendCounter = (2*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:41'!
testHookStore

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter storeCounter |
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.
	storeCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: Stored for: t do: [
			storeCounter := storeCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
		self store: t.	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: storeCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookWrite

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivWriteCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:48'!
testHooks

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:49'!
testHooksWithHalt

	"press proceed for the test to finish"

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

		
	t := TestObject new.
	
	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		self halt.
		receiveCounter := receiveCounter + 1.
	].
	
	Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
		self halt.
		sendCounter := sendCounter + 1.	
	].
	
	Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
		self halt.
		ivReadCounter := ivReadCounter + 1.
	].

	Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
		self halt.
		ivWriteCounter := ivWriteCounter + 1.
	].

	Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
		self halt.
		argCounter := argCounter + 1.
	].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	self assert: receiveCounter = 3.
	self assert: sendCounter = 2.
	self assert: ivReadCounter = 1.
	self assert: ivWriteCounter = 1.
	self assert: argCounter = 1.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:52'!
testHooksWithOverride

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
			reflectogram override: true.
			reflectogram returnValue: 666.
			reflectogram enable.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 666). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/4/2013 12:07'!
testOnInteractionWithFromDebuggingSupport

	|object collection|

	object := Object new.
	collection := OrderedCollection new.
	
	RunTimeDebuggingSupport onInteractionFor: object with: collection do:  '[:r1 :r2 | 
		r1 trigger halt.
		r2 override: true.
		r2 returnValue: r2 defaultAction.
	]'.

	RunTimeDebuggingSupport enableMetaBehaviorForRegistered.
	
	collection add: object.
	
	RunTimeDebuggingSupport removeMetaBehaviorForRegistered.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:53'!
testOneShot

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 1. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:53'!
testOneShotWithPost

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
		receiveCounter := receiveCounter + 1.
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 2. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/10/2013 10:28'!
testSetUnset

	|object|
	
	object := Object new.
	Interpreter enableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object).
	Interpreter disableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object) not.! !

Object subclass: #BenchmarkObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!BenchmarkObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
test
	| a b c |
	a := 1.
	b := 2.
	c := a + b.
	^ c
! !

Object subclass: #Interpreter
	instanceVariableNames: ''
	classVariableNames: 'BaseMetaDict'
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/8/2013 15:04'!
baseMetaDict

	^ BaseMetaDict! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:12'!
disableMetaBehaviorFor: anObject

	"unbind"
	
	self primitiveSetExperimentalBitOf: anObject to: false. ! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 9/2/2013 16:16'!
disableMetaBehaviorForRegistered

	BaseMetaDict keys do: [:key | self disableMetaBehaviorFor: key].! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:11'!
enableMetaBehaviorFor: anObject

	"bind"
	
	self primitiveSetExperimentalBitOf: anObject to: true. ! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 6/22/2013 16:39'!
enableMetaBehaviorForRegistered

	BaseMetaDict keys do: [:key | self enableMetaBehaviorFor: key].! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 6/2/2013 15:59'!
initialize

	super initialize.
	BaseMetaDict := FLLargeIdentityDictionary new.
	Smalltalk specialObjectsArray becomeForward: self newSpecialObjectsArray.! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:13'!
isMetaBehaviorEnabledFor: anObject


	^ self primitiveGetExperimentalBitOf: anObject! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 6/17/2013 11:44'!
newSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	
	| newArray |
	newArray := Array new: 77.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (Smalltalk globals associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: nil. "Was BlockContext"
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (Smalltalk specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 nil 0
							nil 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (Smalltalk specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: Smalltalk compactClassesArray.
	newArray at: 30 put: (Smalltalk specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (Smalltalk specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: 0@0.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (Smalltalk specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (Smalltalk specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (Smalltalk specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((Smalltalk specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (Smalltalk at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (Smalltalk at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (Smalltalk at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (Smalltalk at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (Smalltalk at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (Smalltalk at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (Smalltalk at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (Smalltalk at: #WeakFinalizationList ifAbsent: []).
	newArray at: 59 put: #metaReceived:.
	newArray at: 60 put: #metaRef.
	newArray at: 61 put: #metaReadAt:.
	newArray at: 62 put: #metaWriteAt:Put:.
	newArray at: 63 put: #metaArgIn:for:.
	newArray at: 64 put: self.
	newArray at: 65 put: #metaSend:from:.
	newArray at: 66 put: self baseMetaDict.
	"optimization"
	newArray at: 67 put: self class>>#object:metaReceived:.
	newArray at: 68 put: self class>>#object:metaSend:withReceiver:.
	newArray at: 69 put: self class>>#object:metaReadAt:.
	newArray at: 70 put: self class>>#object:metaWriteAt:Put:.
	newArray at: 71 put: self class>>#object:metaArgIn:withReceiver:.
	
	newArray at: 72 put: #object:metaReceived:.
	newArray at: 73 put: #object:metaSend:withReceiver:.
	newArray at: 74 put: #object:metaReadAt:.
	newArray at: 75 put: #object:metaWriteAt:Put:.
	newArray at: 76 put: #object:metaArgIn:withReceiver:.
	newArray at: 77 put: #object:metaStoreAt:for:.
	^newArray! !

!Interpreter class methodsFor: 'introspection' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex

	<primitive: 73>
	! !

!Interpreter class methodsFor: 'modification' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex put: anotherObject

	<primitive: 74>
	! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/4/2013 17:36'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) argCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:  [Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/4/2013 17:41'!
object: anObject metaReadAt: anIndex

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) readCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject at: anIndex.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: anObject at: anIndex. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: anObject at: anIndex]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject at: anIndex.				
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/4/2013 17:22'!
object: anObject metaReceived: aMessage

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) receiveCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:  [ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/4/2013 17:45'!
object: anObject metaSend: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) sendCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.	
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/17/2013 10:38'!
object: anObject metaStoreAt: anIndex for: anotherObject

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) storeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anotherObject at: anIndex put: anObject
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anotherObject at: anIndex put: anObject
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anotherObject at: anIndex put: anObject]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anotherObject at: anIndex put: anObject.
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/4/2013 17:46'!
object: anObject metaWriteAt: anIndex Put: aValue

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) writeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anObject at: anIndex put: aValue
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject at: anIndex put: aValue
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anObject at: anIndex put: aValue]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject at: anIndex put: aValue
	]. ! !

!Interpreter class methodsFor: 'execution' stamp: 'NickPapoulias 5/5/2013 18:44'!
object: anObject perform: aSelector withArguments: args 

	<primitive: 84>! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 9/2/2013 14:32'!
on: anOperationOrEvent for: anObject do: aCallback

	| implicitMetaObject |
	
	anOperationOrEvent class = ObjectInteraction ifTrue: [
		anOperationOrEvent registerInteractionCallback: aCallback  for: anObject 
	] ifFalse: [
	
		implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
			BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
		].

		anOperationOrEvent registerCallback: aCallback to: implicitMetaObject 
	]! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onArgumentForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject argCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvReadForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject readCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvWriteForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject writeCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageReceivedForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject receiveCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageSendFromObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject sendCallback: aCallback.
! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:47'!
primitiveGetExperimentalBitOf: anObject

	<primitive: 150>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:22'!
primitiveSetExperimentalBitOf: anObject to: aBoolean

	<primitive: 151>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 9/2/2013 16:16'!
removeMetaBehaviorForRegistered

	BaseMetaDict keys do: [:key | self removeMetaBehaviorOf: key].! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 6/2/2013 16:02'!
removeMetaBehaviorOf: anObject

	"unbind and remove implicit meta-object"
	self disableMetaBehaviorFor: anObject.
	BaseMetaDict removeKey: anObject ifAbsent: [].! !

Interpreter subclass: #OptimizedInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/8/2013 14:46'!
disableMetaBehaviorFor: anObject

	"unbind"
	self primitiveSetExperimentalBitOf: anObject to: false. 
	self uninstallMetaBehaviorFor: anObject.! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/8/2013 14:44'!
enableMetaBehaviorFor: anObject

	"bind"
	self installMetaHandlerFor: anObject.
	self primitiveSetExperimentalBitOf: anObject to: true. ! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/8/2013 14:43'!
installMetaHandlerFor: anObject

	|copy implicitMetaObject|
	
	(self object: anObject perform: #class withArguments: #()) = MetaHandler ifTrue: [ ^ anObject].
	copy := anObject copy.
	implicitMetaObject :=  (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]).

	^ anObject becomeForward: 
			(MetaHandler new
				baseObject: copy;
				implicitMetaObject: implicitMetaObject).! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/6/2013 18:44'!
isMetaBehaviorEnabledFor: anObject


	^ self primitiveGetExperimentalBitOf: anObject! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/8/2013 15:26'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := anObject implicitMetaObject argCallback.
	
	callback ifNil: [
		Interpreter primitiveSetExperimentalBitOf: anObject to: true. .
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter primitiveSetExperimentalBitOf: anObject to: true. .
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject baseObject withDefaultAction:  [Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/8/2013 15:27'!
object: anObject metaReadAt: anIndex

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := anObject implicitMetaObject readCallback.
	
	callback ifNil: [
		Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
		^  Interpreter object: anObject baseObject at: anIndex.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
			^   Interpreter object: anObject baseObject at: anIndex. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject baseObject withDefaultAction:   [Interpreter object: anObject baseObject at: anIndex]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
		^ Interpreter object: anObject baseObject at: anIndex.				
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/8/2013 15:29'!
object: anObject metaReceived: aMessage

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := anObject implicitMetaObject receiveCallback.
	
	callback ifNil: [
		Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject baseObject withDefaultAction:  [ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter primitiveSetExperimentalBitOf: anObject to: true. 
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/8/2013 14:53'!
object: anObject metaSend: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := anObject implicitMetaObject sendCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments. 
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.	
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/8/2013 14:54'!
object: anObject metaStoreAt: anIndex for: anotherObject

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := anObject implicitMetaObject storeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anotherObject at: anIndex put: anObject
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anotherObject at: anIndex put: anObject
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anotherObject at: anIndex put: anObject]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anotherObject at: anIndex put: anObject.
	]. ! !

!OptimizedInterpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 10/8/2013 14:55'!
object: anObject metaWriteAt: anIndex Put: aValue

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := anObject implicitMetaObject writeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anObject baseObject at: anIndex put: aValue
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject baseObject at: anIndex put: aValue
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anObject baseObject at: anIndex put: aValue]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject baseObject at: anIndex put: aValue
	]. ! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/8/2013 16:24'!
testResizeFor: anObject
	| class newObject index |
	class := self class.
	class isVariable
		ifTrue: 
			[index := self basicSize.
			newObject := class basicNew: index + 1. "implicit meta-object"
			[index > 0]
				whileTrue: 
					[newObject basicAt: index put: (self basicAt: index).
					index := index - 1].
			newObject basicAt: self basicSize + 1 put:  
				(BaseMetaDict at: anObject ifAbsent: [
					BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
			])]
		ifFalse: [newObject := class basicNew]. "fall back to searching through the base-meta-dict"
	index := class instSize.
	[index > 0]
		whileTrue: 
			[newObject instVarAt: index put: (self instVarAt: index).
			index := index - 1].
	^ newObject! !

!OptimizedInterpreter class methodsFor: 'binding' stamp: 'NickPapoulias 10/8/2013 14:46'!
uninstallMetaHandlerFor: anObject

	^ anObject becomeForward: anObject implicitMetaObject.
			! !

Object subclass: #LazyReifications
	instanceVariableNames: 'interpreterContext'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LazyReifications class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 10:54'!
with: anInterpreterContext

	^ self new interpreterContext: anInterpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext

	^ interpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext: aContext

	interpreterContext := aContext! !

LazyReifications subclass: #Reifications
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:11'!
context

	^ interpreterContext sender ! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
index

	^ interpreterContext tempNamed: #anIndex! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
message

	^ self interpreterContext tempNamed: #aMessage
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
method

	^ self context method! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
receiver

	^ interpreterContext tempNamed: #rcvr
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sender

	^ self context receiver! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sourceRange

	^ self context debuggerMap 
						rangeForPC: self context pc
						contextIsActiveContext: false 
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
trigger

	^ interpreterContext tempNamed: #anObject! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:14'!
value

	^ interpreterContext tempNamed: #aValue! !

Object subclass: #LowLevelImplicitMetaObject
	instanceVariableNames: 'receiveCallback sendCallback readCallback writeCallback argCallback storeCallback'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LowLevelImplicitMetaObject class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/7/2013 17:57'!
new

	^ self basicNew initialize! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback
	^ argCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback: anObject
	argCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/2/2013 15:16'!
initialize 

	super initialize.

	"self receiveCallback: [:reifications :override | ].

	self sendCallback: [:reifications :override | ].

	self readCallback: [:reifications :override | ].

	self writeCallback: [:reifications :override | ].
	
	self	 argCallback: [:reifications :override | ].
		"! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback
	^ readCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback: anObject
	readCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback
	^ receiveCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback: anObject
	receiveCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback
	^ sendCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback: anObject
	sendCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/17/2013 11:29'!
storeCallback
	^ storeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/17/2013 11:29'!
storeCallback: anObject
	storeCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback
	^ writeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback: anObject
	writeCallback := anObject! !

Object subclass: #MetaBenchmark
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/15/2013 14:47'!
benchmark: aBlock times: anInteger
	"synced with Bifrost's bench"
	| runs average deviation |
	aBlock value; value. "this does nothing in our case possible bifrost uses it
	to install methodWrappers the first time this is run ?"
	runs := [ 
		(1 to: anInteger) 
			collect: [ :each | aBlock timeToRun ] ]
		valueUnpreemptively.
	average := runs average asFloat.
	deviation := ((runs 
		detectSum: [ :each | (each - average) * (each - average) ]) 
			/ runs size) sqrt.
	^ Array with: average with: deviation! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
message
	"self new message" 
	
	|  object  |
	object := BenchmarkObject new.
	^ self benchmark: [ 1 to: 1000000 do: [ :each | object test ] ] times: 100.! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/5/2013 11:57'!
messageMetaJumpWithHaltOnly
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	HaltInterpreter initialize.
	 
	HaltInterpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	HaltInterpreter removeMetaBehaviorOf: object.
	
	Interpreter initialize.
	
	^ result! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 17:10'!
messageMetaSend
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter on: MessageReceived for: object do: [].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 17:51'!
messageMetaSendWithReifications
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter on: MessageReceived for: object do: [:reifications :reflectogram | ].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

Object subclass: #MetaHandler
	instanceVariableNames: 'baseObject implicitMetaObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
baseObject
	^ baseObject! !

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
baseObject: anObject
	baseObject := anObject! !

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
implicitMetaObject
	^ implicitMetaObject! !

!MetaHandler methodsFor: 'accessing' stamp: 'NickPapoulias 10/7/2013 16:45'!
implicitMetaObject: anObject
	implicitMetaObject := anObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit
	^ self primitiveGetExperimentalBit ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit: aBoolean
	self primitiveSetExperimentalBitTo: aBoolean! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/8/2013 15:34'!
identityHash
	"Answer a SmallInteger whose value is related to the receiver's identity.
	This method must not be overridden, except by SmallInteger.
	
	Do not override."

	^ (Interpreter object: self perform: #basicIdentityHash withArguments: {}) bitShift: 18! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
isExperimentalBitOn
	^ self experimentalBit! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaArgIn: aMessage for: anObject

	^ Interpreter object: anObject metaArgIn: aMessage withReceiver: self! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReadAt: anIndex

	^ Interpreter object: self metaReadAt: anIndex! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReceived: aMessage

	^ Interpreter object: self metaReceived: aMessage ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:39'!
metaRef

	self halt! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaSend: aMessage from: anObject

	^ Interpreter object: anObject metaSend: aMessage withReceiver: self.! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaWriteAt: anIndex Put: anotherObject

	^ Interpreter object: self metaWriteAt: anIndex Put: anotherObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveGetExperimentalBit

	<primitive: 150>
	self primitiveFailed! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveSetExperimentalBitTo: aBoolean

	<primitive: 151>
	self primitiveFailed! !

Object subclass: #Operation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #AsArgument
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!AsArgument class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject argCallback: aCallback ! !

Operation subclass: #AsReturnValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #CompositeOperation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeOperation subclass: #ObjectCreation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeOperation subclass: #ObjectInteraction
	instanceVariableNames: 'interactionObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!ObjectInteraction class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 18:31'!
with: anotherObject

	^ self new with: anotherObject ! !

!ObjectInteraction methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:51'!
registerInteractionCallback: aCallback for: anObject

	Interpreter on: AsArgument for: anObject do: [:reifications :reflectogram | 
		reifications receiver = interactionObject ifTrue: [aCallback value: reifications value: reflectogram]	
	].
	
	Interpreter on: MessageReceived for: anObject  do: [:reifications :reflectogram | 
		(reifications message arguments includes: interactionObject) ifTrue: [aCallback value: reifications value: reflectogram]	
	].
! !

!ObjectInteraction methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/2/2013 14:03'!
with: anotherObject

	interactionObject := anotherObject! !

Operation subclass: #IvRead
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvRead class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject readCallback: aCallback ! !

Operation subclass: #IvWrite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvWrite class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:23'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject writeCallback: aCallback ! !

Operation subclass: #MessageReceived
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageReceived class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:24'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject receiveCallback: aCallback ! !

Operation subclass: #MessageSent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageSent class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:24'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject sendCallback: aCallback ! !

!Operation class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:15'!
registerCallback: aCallback to: anImplicitMetaObject

	self subclassResponsibility ! !

Operation subclass: #Referenced
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #Stored
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Stored class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject storeCallback: aCallback ! !

Object subclass: #Reflectogram
	instanceVariableNames: 'trigger returnValue defaultAction override'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:18'!
for: anObject withDefaultAction: aBlock

	^ self basicNew initializeWithTrigger: anObject andAction: aBlock! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:56'!
cascade: aBlock

	| ret |
	Interpreter enableMetaBehaviorFor: trigger.
	ret := aBlock value.
	Interpreter disableMetaBehaviorFor: trigger.
	^ ret.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:07'!
defaultAction

	^ defaultAction value! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
disable 

	Interpreter disableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:56'!
enable

	Interpreter enableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:45'!
initializeWithTrigger: anObject andAction: aBlock

	trigger := anObject.
	defaultAction := aBlock.
	override := false.
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
isEnabled 

	^ Interpreter isMetaBehaviorEnabledFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:44'!
override

	^ override ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:43'!
override: aBoolean

	override := aBoolean ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 18:12'!
remove

	Interpreter removeMetaBehaviorOf: trigger.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue

	^ returnValue ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue: aValue

	returnValue := aValue! !

Object subclass: #TestObject
	instanceVariableNames: 'x'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 16:04'!
testInteraction: anObject 

	anObject class.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 15:58'!
testPassingArgument: anObject ! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/9/2013 19:38'!
testSenderHook

	Transcript show: 'this message should be caught'; cr.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:01'!
x

	^ x! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/5/2013 19:02'!
x: anObject 

	x := anObject ! !
Interpreter initialize!
